##############################################################################
# This file plots the pathogen concentration and the absorbed load for the
# close contact scenario. The plots are saved as pdf and eps.
# Date: 2022-07-14
# Author: Simon Rahn
##############################################################################

import pandas as pd
import numpy as np
import os

import sys
# insert at 1, 0 is the script path (or '' in REPL)
sys.path.insert(1, '..\\')
from plotSettingsPlosOne import *


def read_csv_as_dataframe(filepath, sep=" "):
    dataframe = pd.read_csv(filepath, sep=sep)
    return dataframe


if __name__ == '__main__':
    pathToOutputDir = "..\\..\\output\\closeContact\\"
    scenario = "closeContact"

    # define model settings
    simTimeStepLengthInSec = 0.4

    # common plot settings
    figsize = [figWidth, figHeight]
    adjustParam = {'left' : 0.13, 'bottom' : 0.23, 'right' : 0.964, 'top' : 0.78, 'wspace' : 0.543, 'hspace' : 0.195}
    xLabel = "Time in s"
    xlim = [0, 900]
    xticks = np.linspace(start=xlim[0], stop=xlim[1], num=4, endpoint=True)

    # common analysis settings
    auxTime = 600

    # common save options
    saveFig = True
    figDirPath = os.path.join(".")

    ##########################
    # pathogen concentration
    ##########################
    file = "pathogenConcentration.txt"
    data = read_csv_as_dataframe(os.path.join(pathToOutputDir, scenario, file))
    key = 'gridConcentration-PID6'

    # figure settings
    yLabel = "Concentration in $\\frac{\\mathrm{particles}}{\\mathrm{m}^3}$"
    ylim = [0, 100000]

    # filter data: we only want to analyze the pathogen concentration at coordinate (3, 3)
    filter1 = data['x'] == 3
    filter2 = data['y'] == 3
    data = data.where(filter1 & filter2).dropna()

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize, dpi=dpi)
    ax1.plot(data['timeStep'].where(data['timeStep'] * simTimeStepLengthInSec <= auxTime).dropna() * simTimeStepLengthInSec,
            data[key].where(data['timeStep'] * simTimeStepLengthInSec <= auxTime).dropna())
    ax1.plot(data['timeStep'].where(data['timeStep'] * simTimeStepLengthInSec > auxTime).dropna() * simTimeStepLengthInSec,
            data[key].where(data['timeStep'] * simTimeStepLengthInSec > auxTime).dropna(), color='tab:gray', ls='--')

    ax1.set_xlabel(xLabel)
    ax1.set_xlim(xlim)
    ax1.set_xticks(xticks)

    ax1.set_ylabel(yLabel)
    ax1.set_ylim(ylim)
    ax1.set_yticks([0, 20000, 40000, 60000, 80000, 100000])
    ax1.ticklabel_format(axis="y", style="sci", scilimits=(0,0))
    ax1.ticklabel_format(useMathText=True)

    ax1.spines['top'].set_visible(False)
    ax1.spines['right'].set_visible(False)

    ax1.set_title(label='A', fontweight='bold', x=0, y=1.15, fontsize=captionFontSize)


    ##########################
    # absorbed pathogen load
    ##########################
    file = "pedestrianPathogenLoad.txt"
    data = read_csv_as_dataframe(os.path.join(pathToOutputDir, scenario, file))
    key = 'absorbedPathogenLoad-PID5'

    filter1 = (data['timeStep'] * simTimeStepLengthInSec > auxTime - simTimeStepLengthInSec / 2)
    filter2 = (data['timeStep'] * simTimeStepLengthInSec < auxTime + simTimeStepLengthInSec / 2)
    filter3 = (data[key] > 0)
    auxData = data.where(filter1 & filter2 & filter3).dropna()

    # figure settings
    yLabel = "Number of absorbed\npathogen particles"
    ylim = [0, 6000]

    # filter data: we only want to analyze the pathogen concentration at coordinate (3, 3)
    filter1 = data['pedestrianId'] == 1
    data = data.where(filter1).dropna()

    # fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    ax2.plot(data['timeStep'].where(data['timeStep'] * simTimeStepLengthInSec <= auxTime).dropna() * simTimeStepLengthInSec,
            data[key].where(data['timeStep'] * simTimeStepLengthInSec <= auxTime).dropna())
    ax2.plot(data['timeStep'].where(data['timeStep'] * simTimeStepLengthInSec > auxTime).dropna() * simTimeStepLengthInSec,
            data[key].where(data['timeStep'] * simTimeStepLengthInSec > auxTime).dropna(), color='tab:gray', ls='--')
    ax2.plot(np.multiply([0, auxData.iloc[0]['timeStep'], auxData.iloc[0]['timeStep']], simTimeStepLengthInSec),
            [auxData.iloc[0][key], auxData.iloc[0][key], 0],
            ls='dotted',
            color='k')

    ax2.set_xlabel(xLabel)
    ax2.set_xlim(xlim)
    ax2.set_xticks(xticks)

    ax2.set_ylabel(yLabel)
    ax2.set_ylim(ylim)
    ax2.set_yticks([0, 2000, 3200, 4000, 6000])
    ax2.ticklabel_format(axis="y", style="scientific", scilimits=(0,0))
    ax2.ticklabel_format(useMathText=True)

    ax2.spines['top'].set_visible(False)
    ax2.spines['right'].set_visible(False)

    ax2.set_title(label='B', fontweight='bold', x=0, y=1.15, fontsize=captionFontSize)

    plt.subplots_adjust(left=adjustParam['left'], bottom=adjustParam['bottom'], right=adjustParam['right'], top=adjustParam['top'], wspace=adjustParam['wspace'], hspace=adjustParam['hspace'])

    fig.show()

    # save figure
    if (saveFig):
        figFileName = "pathogenConcentrationAndLoad"
        plt.savefig(os.path.join(figDirPath, figFileName + ".pdf"))
        plt.savefig(os.path.join(figDirPath, figFileName + ".tif"))

    print('finished')
