##############################################################################
# This file evaluates the average distances between the agents in a queue.
# Date: 2022-07-14
# Author: Simon Rahn
##############################################################################
import math
import pickle

import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import os


# insert at 1, 0 is the script path (or '' in REPL)
import sys
sys.path.insert(1, '..\\')
from plotSettingsPlosOne import *


def get_pos_by_ped_id(data, id):
    dataPed = data[data['pedestrianId']==id]
    if dataPed.values.shape[0] < 1:
        pos = []
    else:
        dataPedTmax = dataPed[dataPed['simTime'] == dataPed.simTime.max()]
        pos = [dataPedTmax.get('startX-PID1').values.tolist()[0], dataPedTmax.get('startY-PID1').values.tolist()[0]]
    return pos


def get_euclidean_distance(pos1, pos2):
    return math.sqrt(math.pow((pos1[0]-pos2[0]), 2) + math.pow((pos1[1]-pos2[1]), 2))


if __name__ == '__main__':

    pathToOutputSubdir = "..\\..\\output\\queue\\parameter_variation\\"

    # read data
    dumpedFileName = "distances.pkl"
    dumpedFileName2 = "contactTimesToInfectiousAgent.pkl"
    dumpedFilePath = os.path.join(pathToOutputSubdir, dumpedFileName)
    dumpedFilePath2 = os.path.join(pathToOutputSubdir, dumpedFileName2)

    # initialize variables
    deltaT = 1  # time step width in which the results are evaluated
    startT = 0
    stopT = 1400    # simEndTime
    nSteps = int((stopT - startT)/deltaT+1)
    timeSteps = np.linspace(startT,stopT,num=nSteps)
    distanceSet = np.zeros([6, nSteps - 1])
    distanceToInfectiousAgentTimesSet = np.zeros([5, 10])
    infectiousAgentId = 5

    if os.path.isfile(dumpedFilePath) & os.path.isfile(dumpedFilePath2):
        with open(dumpedFilePath,'rb') as f:
            distanceSet = pickle.load(f)

        with open(dumpedFilePath2,'rb') as f:
            distanceToInfectiousAgentTimesSet = pickle.load(f)

    else:
        k = 0   # index variable for number of subdirs

        subdirs = os.listdir(pathToOutputSubdir)
        for subdir in subdirs:
            if os.path.isdir(os.path.join(pathToOutputSubdir, subdir)):
                parameterSet = subdir[-3:]
                subsubdirs = os.listdir(os.path.join(pathToOutputSubdir, subdir))
                file = "postvis.traj"
                pathToOutputFile = os.path.join(pathToOutputSubdir, subdir, subsubdirs[0], file)

                data = pd.read_csv(pathToOutputFile, sep=" ")

                # initialize temp variables
                distance = np.zeros([9, nSteps - 1])    # stores distances between agent i and agent i+1, i = 1,...,10
                distanceToInfectiousAgent = np.zeros([10, nSteps - 1])  # stores distances between agent i and agent <infectiousAgentId>, i = 1,...,10

                for t in range(0, nSteps - 1):
                    tStepMin = timeSteps[t]
                    tStepMax = timeSteps[t+1]
                    dataf1 = data[(data['simTime']>=tStepMin) & (data['simTime']<tStepMax)]
                    positions = []

                    for id in range(1,11):
                        positions.append(get_pos_by_ped_id(dataf1, id))

                    for i in range(0, 9):
                        # calculate and store of distances in the queue between agent i and agent i+1
                        if (len(positions[i]) > 0) & (len(positions[i + 1]) > 0):
                            distance[i, t] = get_euclidean_distance(positions[i], positions[i + 1])
                        else:
                            distance[i, t] = np.nan

                    for i in range(0, 10):
                        # calculate and store of distances in the queue between infectious agent and agent i
                        if ((len(positions[infectiousAgentId-1]) > 0) & (len(positions[i]) > 0)):
                            distanceToInfectiousAgent[i, t] = get_euclidean_distance(positions[infectiousAgentId-1], positions[i])
                        else:
                            distanceToInfectiousAgent[i, t] = np.nan

                    if k == 0:
                        distanceSet[0, t] = timeSteps[t]
                    print(timeSteps[t])

                distanceSet[k + 1, :] = np.nanmean(distance, axis=0)

                # calculate period of time during which this condition holds: distances < threshold
                threshold = 1.5
                distanceToInfectiousAgentTimesSet[k, :] = (distanceToInfectiousAgent < threshold).sum(axis=1) * deltaT

                k = k+1

        with open(dumpedFilePath,'wb') as f:
            pickle.dump(distanceSet, f)

        with open(dumpedFilePath2, 'wb') as f:
            pickle.dump(distanceToInfectiousAgentTimesSet, f)

    # common plot options
    savefig = True
    xlim = [0, 1150]
    ylim = [0.5, 3]
    labels = [0.5, 1.0, 1.5, 2.0, 2.5]
    # labels = ['$P=0.5$', '$P=1.0$','$P=1.5$','$P=2.0$','$P=2.5$']
    labels = ['A','B','C','D','E']
    adjustParam = {'left': 0.22, 'bottom': 0.214, 'right': 0.976, 'top': 0.964}

    color = plt.cm.Blues_r(np.linspace(0.2,0.8,5))
    mpl.rcParams['axes.prop_cycle'] = mpl.cycler(color=color)
    figDirPath = os.path.join(".")

    figsize = [figWidth/1.95, figHeight + 0.1]

    ############
    # plot 2
    ############
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    distanceSetClean = distanceSet.copy()
    timeCutoffIdx = 51 # ignore first timeSteps during which the queue forms
    distanceSetClean[1:, :timeCutoffIdx] = np.nan

    for k in range(1,6):
        ax.plot(distanceSetClean[0, :1151], distanceSetClean[k, :1151], label=labels[k-1], linewidth=1)
        mean = np.nanmean(distanceSetClean[k, :].transpose())
        ax.plot(xlim, [mean, mean], label="_nolegend_", color='k', linestyle=':', linewidth=0.75)
        # annotations
        # ax.text(xlim[0]+5, mean + 0.025, f"{mean:.1f}", fontsize='x-small')
        # ax.text(xlim[0]-5, mean, f"{mean:.2f}", fontsize='x-small', horizontalalignment='right', verticalalignment ='center')
        # ax.text(distanceSetClean[0, 1010] + 0.05, distanceSetClean[k, 1010], labels[k-1], horizontalalignment='left', verticalalignment ='bottom')

    ax.legend(framealpha=1, ncol=5, columnspacing=0.75, labelspacing=0.2,handlelength=0.5,handletextpad=0.2)
    ax.set_xlabel('Time in s')
    ax.set_ylabel('Mean distance $\overline{d_p}$ in m')
    ax.set_xlim(xlim)
    ax.set_xticks([0, 500, 1000])
    ax.set_ylim(ylim)
    plt.subplots_adjust(left=adjustParam['left'], bottom=adjustParam['bottom'], right=adjustParam['right'], top=adjustParam['top'])

    # save file
    if savefig:
        figFileName = "linePlot_time-distances_mean"
        plt.savefig(os.path.join(figDirPath, figFileName + ".pdf"))
        plt.savefig(os.path.join(figDirPath, figFileName + ".eps"))
        plt.savefig(os.path.join(figDirPath, figFileName + ".tif"))
    else:
        plt.show()

    plt.close()

    print('done')

