# -----------------------------------------------------------
# CovidSim - TransmissionModel validation
#
# This script generates a parametrized scenario file from a template (see pathToScenarioTemplateFile). The scenario
# models a Covid-19 superspreading event recorded in [hamner-2020-life] and [miller-2020-life]. It is used for the
# validation of Vadere's TransmissionModel.
#
# [hamner-2020-life] Hamner L, Dubbel P, Capron I, et al. High SARS-CoV-2 Attack Rate Following Exposure at a
# Choir Practice — Skagit County, Washington, March 2020. MMWR Morb Mortal Wkly Rep 2020;69:606–610.
# DOI: http://dx.doi.org/10.15585/mmwr.mm6919e6external
#
# [miller-2020-life] Shelly L. Miller, William W Nazaroff, Jose L. Jimenez, Atze Boerstra, Giorgio Buonanno,
# Stephanie J. Dancer, Jarek Kurnitski, Linsey C. Marr, Lidia Morawska, and Catherine Noakes. Transmission
# of SARS-CoV-2 by inhalation of respiratory aerosol in the skagit valley chorale superspreading event.
# 31(2):314–323. https://doi.org/10.1111/ina.12751.
#
# Author: Simon Rahn
# Date: 2022-07-14
# -----------------------------------------------------------

import numpy as np
import random as rd
import json
import pandas as pd

from imports import *


def generate_scenario_elements(nRows, nColumns, xStart, yStart, xSpace, ySpace, elementIdStart, elementWidth,
                               elementHeight, elementDict, nMaxElements, waitingTime=0):
    # start at upper left corner, complete a row, continue with the next row below the previous one

    global TARGET_LISTS

    scenarioElementDictList = list()
    elementId = elementIdStart
    elementPosY = yStart + ySpace  # in the first loop ySpace is subtracted again
    idxElements = 0
    idxTargetId = 0

    for rows in range(nRows):
        elementPosX = xStart  # reset x position for each new row
        elementPosY = elementPosY - ySpace

        for column in range(nColumns):
            shapeDict = {"x": elementPosX,
                         "y": elementPosY,
                         "width": elementWidth,
                         "height": elementHeight,
                         "type": "RECTANGLE"}

            elementDict['id'] = elementId
            elementDict['shape'] = shapeDict

            if 'waitingTime' in elementDict:
                elementDict['waitingTime'] = waitingTime

            if 'targetIds' in elementDict:
                targets = []
                for i in range(len(TARGET_LISTS)):
                    targets.append(TARGET_LISTS[i][idxTargetId])

                idxTargetId = idxTargetId + 1

                elementDict['targetIds'] = targets

            scenarioElementDictList.append(elementDict.copy())

            idxElements = idxElements + 1
            if idxElements == nMaxElements:
                return scenarioElementDictList

            elementPosX = elementPosX + xSpace
            elementId = elementId + 1

    return scenarioElementDictList


if __name__ == '__main__':
    # coordinates:
    # X -> horizontal / width / rows
    # Y -> vertical / height / depth / columns
    #
    # units: all values in SI units (m, s)
    #
    # assumptions are commented with [assumption]
    # information taken from [hamner-2020-life] are referenced

    # total number of agents
    nAgents = 61  # [hamner-2020-life]

    # Usage of seeds / random numbers in this script
    createNewTargetList = False
    # False: load previously generated list of targets
    # True: generate new seeds (this may create different random numbers on different computers / python versions
    # / etc., save that new list
    dumpedFileName = "targetList"
    dumpedFilePath = os.path.join(pathToTargets, dumpedFileName)

    if createNewTargetList:
        # initialize dictionary that saves all targets for each agent
        # targetDictionary = dict.fromkeys(['Seed', 'Source ID', 'Target 1', 'Target 2', 'Target 3', 'Target 4', 'Target 5'])
        targetDictionary = {'Seed': [],
                            'Source ID': [],
                            'Target 1': [],
                            'Target 2': [],
                            'Target 3': [],
                            'Target 4': [],
                            'Target 5': []}
    else:
        targetDataFrame = pd.read_pickle(dumpedFilePath + '.pkl')

    # generate random number to assure reproducibility
    for seed in range(startSeed, nSeeds + startSeed):
        rd.seed(seed)

        # topography
        boundingBoxWidth = 0.5

        # sources area
        nSources = nAgents
        nSourcesColumns = 6
        nSourcesRows = int(np.ceil(nSources / nSourcesColumns))
        sourceWidth = 0.41
        sourceHeight = 0.41
        sourceSpacingX = 0.4
        sourceSpacingY = 0.4
        sourcesAreaBoundingBoxWidth = 1
        sourceIdCounter = 1  # start source ids with sourceIdCounter

        sourcesAreaWidth = sourcesAreaBoundingBoxWidth * 2 + nSourcesColumns * sourceWidth + (
                nSourcesColumns - 1) * sourceSpacingX
        sourcesAreaHeight = sourcesAreaBoundingBoxWidth * 2 + nSourcesRows * sourceHeight + (
                nSourcesRows - 1) * sourceSpacingY

        # wall thickness
        wallThickness = 0.5

        # targets
        targetIdCounter = 1001  # start target ids with targetIdCounter

        # chairs
        seatWidthX = 0.4  # [assumption]
        seatDepthY = 0.4  # [assumption]

        targetWidth = 0.4
        targetHeight = 0.4
        relativeTargetPositionX = seatWidthX / 2 - targetWidth / 2
        relativeTargetPositionY = - targetHeight

        # stage
        nChairRows = 6  # [hamner-2020-life]
        chairRowSpacing = 1.4 - seatDepthY  # [miller-2020-life] (distance between two chairs; center to center 1.4 m)

        nChairColumns = 10
        chairColumnSpacing = 0.75 - seatDepthY # [miller-2020-life] horizontal spacing between chairs (center to center 0.75 m);

        stagesSpacingX = 1.2    # [assumption based on total room width ~18m [miller-2020-life]] horizontal spacing between left and right stage / width of aisle
        stageWallSpacingX = 1.2  # [assumption based on total room width ~18m [miller-2020-life]] horizontal spacing between stages and walls

        stageWallSpacingYLower = 2.5  # [assumption based on total room depth ~9m [miller-2020-life]] vertical spacing between stages and walls
        stageWallSpacingYUpper = 1  # [assumption based on total room depth ~9m [miller-2020-life]] vertical spacing between stages and walls

        stagesWidth = nChairColumns * seatWidthX + (nChairColumns - 1) * chairColumnSpacing
        stagesHeight = nChairRows * seatDepthY + (nChairRows - 1) * chairRowSpacing

        # first chair within a stage is in the upper left corner, chair id increases column-wise then row-wise
        # X, Y position defines the upper left corner of the chair -> in Vadere, the position would be X, (Y - seatDepth)
        stageLeftX = boundingBoxWidth + sourcesAreaWidth + wallThickness + stageWallSpacingX
        stageLeftY = boundingBoxWidth + wallThickness + stagesHeight + stageWallSpacingYLower
        stageLeftFirstId = targetIdCounter

        stageRightX = stageLeftX + stagesWidth + stagesSpacingX
        stageRightY = stageLeftY
        stageRightFirstId = stageLeftFirstId + nChairRows * nChairColumns

        # benches, the exact topography is not relevant since the index case remains in the
        nBenchesOccupants = 18  # [miller-2020-life]
        nBenchesColumns = 10  # [assumption]
        benchSeatColumnSpacing = chairColumnSpacing # [assumption]

        nBenchesRows = 3 # [assumption] or int(np.ceil(nBenchesOccupants / nBenchesColumns))
        benchSeatRowSpacing = chairRowSpacing

        benchesWallSpacingX = stageWallSpacingX  # [assumption]
        benchesWallSpacingY = stageWallSpacingYLower  # [assumption]

        benchesWidth = nBenchesColumns * seatWidthX + (nBenchesColumns - 1) * benchSeatColumnSpacing
        benchesHeight = nBenchesRows * seatDepthY + (nBenchesRows - 1) * benchSeatRowSpacing

        benchesX = stageRightX + stagesWidth + stageWallSpacingX + wallThickness + benchesWallSpacingX
        benchesY = boundingBoxWidth + wallThickness + benchesWallSpacingY + benchesHeight
        benchesFirstId = stageRightFirstId + nChairRows * nChairColumns

        # room dimensions
        largeRoomWidth = 2 * stageWallSpacingX + 2 * stagesWidth + stagesSpacingX
        largeRoomHeight = stageWallSpacingYLower + stageWallSpacingYUpper + stagesHeight
        largeRoomX = boundingBoxWidth + sourcesAreaWidth + wallThickness  # X position of bottom left corner of large room
        largeRoomY = boundingBoxWidth + wallThickness

        smallRoomWidth = benchesWidth + 2 * benchesWallSpacingX
        smallRoomHeight = benchesHeight + 2 * benchesWallSpacingY

        # practice phases: in large and small rooms
        practiceTime = 45 * 60  # [hamner-2020-life] but simplified to 45 min instead of practice 40 min, 50min, 45 min

        # break targets -> geometry [assumption]
        nBreakTargets = 16  # [assumption] -> distribute agents in groups across the room during break
        nRestroomTargets = 1
        nBreakTargetColumns = 5
        nBreakTargetRows = int(np.ceil(nBreakTargets / nBreakTargetColumns))
        breakTargetWidth = 1
        breakTargetHeight = 1
        breakTargetSpacingX = (largeRoomWidth - breakTargetWidth * nBreakTargetColumns) / (nBreakTargetColumns + 1)
        breakTargetSpacingY = (chairRowSpacing + seatDepthY) * 2

        # break phase: intermediate targets during break (groups distributed in large room)
        nMaxAgentsPerBreakTarget = 4  # [miller-2020-life] groups of 3-4
        nMaxAgentsInRestroom = 6    # [assumption based on miller-2020-life]
        breakTime = 10 * 60  # [miller-2020-life] ([hamner-2020-life] 15 min period of break)
        nPassedBreakTargets = 1  # [assumption] increase nTargetsDuringBreak to introduce more movement;
        # number of targets that an agent passes during the break
        breakTimePerTarget = breakTime / nPassedBreakTargets
        breakTargetsFirstId = benchesFirstId + nBenchesRows * nBenchesColumns
        restroomTargetsFirstId = breakTargetsFirstId + nBreakTargets

        # final target
        finalTargetId = restroomTargetsFirstId + nRestroomTargets

        # topography
        boundsX = 2 * boundingBoxWidth + sourcesAreaWidth + 3 * wallThickness + largeRoomWidth + smallRoomWidth
        boundsY = 2 * boundingBoxWidth + largeRoomHeight + 2 * wallThickness

        # generate scenario elements
        # targetIds for each source are drawn from lists of random numbers; targetIds at index 0 of each list belong to the
        # source which spawns an infectious agent

        # [assumption] nSkippedRows last rows of chairs are not occupied
        nSkippedRows = 0

        leftStageIds = range(stageLeftFirstId + nSkippedRows * nChairColumns, stageRightFirstId)
        rightStageIdsPracticeTime1 = range(stageRightFirstId + nSkippedRows * nChairColumns, benchesFirstId)

        # positions of agents during practice session 2 in the large room
        # a) [hamner-2020-life] agents move close together -> no free seats between agents (limited number of ids in
        # rightStageIdsPracticeTime2, benchesIds)
        rightStageIdsPracticeTime2 = range(benchesFirstId - (nAgents - nBenchesOccupants),
                                           benchesFirstId)
        # b) [miller-2020-life] agents move to right stage but distribute over all 60 seats
        rightStageIdsPracticeTime2 = range(stageRightFirstId,
                                           benchesFirstId)

        benchesIds = range(benchesFirstId, breakTargetsFirstId)

        breakIds = range(breakTargetsFirstId, breakTargetsFirstId + nBreakTargets)  # [assumption]
        restroomIds = range(restroomTargetsFirstId, restroomTargetsFirstId + nRestroomTargets)

        # generate lists
        if createNewTargetList:
            LIST_1 = list(leftStageIds) + list(rightStageIdsPracticeTime1)
            LIST_2 = list(rightStageIdsPracticeTime2) + list(benchesIds)
            LIST_3a = list(breakIds) * nMaxAgentsPerBreakTarget
            LIST_3b = list(restroomIds) * nMaxAgentsInRestroom      # create separate list for restrooms (during break)
            LIST_3 = LIST_3a[0:nAgents - len(LIST_3b)] + LIST_3b    # remove breakTargetIds that are too many (len(LIST_3) should equal nAgents)
            LIST_3 = LIST_3
            LIST_5 = []
            LIST_6 = [finalTargetId] * nAgents
            # randomize agents' seating positions
            rd.shuffle(LIST_1)
            rd.shuffle(LIST_2)

            # define the infectious agent's positions in practice session 2
            # i.e. move a random number from the interval [a, b] (-> randint(a, b)) to index
            # infectiousSourceIdx in LIST_2
            pos2 = rd.randint(rightStageIdsPracticeTime1[0], rightStageIdsPracticeTime1[-1])    # infectious agent is placed in the right stage of room L during practice session 2
            pos3 = restroomTargetsFirstId
            infectiousSourceIdx = 0
            LIST_2.insert(infectiousSourceIdx, LIST_2.pop(LIST_2.index(pos2)))

            # merge all target lists to common list
            TARGET_LISTS = [LIST_1, LIST_2]
            for idx in range(nPassedBreakTargets):
                rd.shuffle(LIST_3)
                LIST_3.insert(infectiousSourceIdx, LIST_3.pop(LIST_3.index(pos3)))
                TARGET_LISTS.append(LIST_3.copy())
            TARGET_LISTS.append(
                LIST_1)  # add LIST_1 once again, because agents return to their original seats during practice phase 3 (after the break)
            if LIST_5 != []:
                TARGET_LISTS.append(LIST_5)
            TARGET_LISTS.append(LIST_6)

            for i in range(0, nAgents):
                targetDictionary['Seed'].append(seed)
                targetDictionary['Source ID'].append(i + 1)
                targetDictionary['Target 1'].append(TARGET_LISTS[0][i])
                targetDictionary['Target 2'].append(TARGET_LISTS[1][i])
                targetDictionary['Target 3'].append(TARGET_LISTS[2][i])
                targetDictionary['Target 4'].append(TARGET_LISTS[3][i])
                targetDictionary['Target 5'].append(TARGET_LISTS[4][i])

        else:
            # load target lists
            targetDataFrameSeed = targetDataFrame[targetDataFrame['Seed'] == seed]

            TARGET_LISTS = [targetDataFrameSeed['Target 1'].to_list(),
                            targetDataFrameSeed['Target 2'].to_list(),
                            targetDataFrameSeed['Target 3'].to_list(),
                            targetDataFrameSeed['Target 4'].to_list(),
                            targetDataFrameSeed['Target 5'].to_list()]


        # iterate through sources
        sourceDict = {"id": 0,
                      "shape": "",
                      "interSpawnTimeDistribution": "org.vadere.state.scenario.ConstantDistribution",
                      "distributionParameters": [1.0],
                      "spawnNumber": 1,
                      "maxSpawnNumberTotal": -1,
                      "startTime": 0,
                      "endTime": 0,
                      "spawnAtRandomPositions": False,
                      "spawnAtGridPositionsCA": False,
                      "useFreeSpaceOnly": True,
                      "targetIds": [],
                      "groupSizeDistribution": [1.0],
                      "dynamicElementType": "PEDESTRIAN",
                      "attributesPedestrian": None
                      }

        sourceList = generate_scenario_elements(nSourcesRows,
                                                nSourcesColumns,
                                                boundingBoxWidth + sourcesAreaBoundingBoxWidth,
                                                boundsY - boundingBoxWidth - sourcesAreaBoundingBoxWidth - sourceHeight,
                                                sourceSpacingX + sourceWidth,
                                                sourceSpacingY + sourceHeight,
                                                sourceIdCounter,
                                                sourceWidth,
                                                sourceHeight,
                                                sourceDict,
                                                nSources)

        # iterate through targets
        targetList = list()
        targetDict = {"id": 0,
                      "absorbing": False,
                      "shape": "",
                      "waitingTime": 0,
                      "waitingTimeYellowPhase": 0.0,
                      "parallelWaiters": 0,
                      "individualWaiting": True,
                      "deletionDistance": 0.1,
                      "startingWithRedLight": False,
                      "nextSpeed": -1.0}

        stageLeftTargets = generate_scenario_elements(nChairRows,
                                                      nChairColumns,
                                                      stageLeftX + relativeTargetPositionX,
                                                      stageLeftY + relativeTargetPositionY,
                                                      chairColumnSpacing + seatWidthX,
                                                      chairRowSpacing + seatDepthY,
                                                      stageLeftFirstId,
                                                      targetWidth,
                                                      targetHeight,
                                                      targetDict,
                                                      nChairRows * nChairColumns,
                                                      practiceTime)

        stageRightTargets = generate_scenario_elements(nChairRows,
                                                       nChairColumns,
                                                       stageRightX + relativeTargetPositionX,
                                                       stageRightY + relativeTargetPositionY,
                                                       chairColumnSpacing + seatWidthX,
                                                       chairRowSpacing + seatDepthY,
                                                       stageRightFirstId,
                                                       targetWidth,
                                                       targetHeight,
                                                       targetDict,
                                                       nChairRows * nChairColumns,
                                                       practiceTime)

        benchesTargets = generate_scenario_elements(nBenchesRows,
                                                    nBenchesColumns,
                                                    benchesX + relativeTargetPositionX,
                                                    benchesY + relativeTargetPositionY,
                                                    benchSeatColumnSpacing + seatWidthX,
                                                    benchSeatRowSpacing + seatDepthY,
                                                    benchesFirstId,
                                                    targetWidth,
                                                    targetHeight,
                                                    targetDict,
                                                    nBenchesRows * nBenchesColumns,
                                                    practiceTime)

        breakTargets = generate_scenario_elements(nBreakTargetRows,
                                                  nBreakTargetColumns,
                                                  largeRoomX + breakTargetSpacingX,
                                                  stageLeftY + relativeTargetPositionY - chairRowSpacing,
                                                  breakTargetSpacingX + breakTargetWidth,
                                                  breakTargetSpacingY,
                                                  breakTargetsFirstId,
                                                  breakTargetWidth,
                                                  breakTargetHeight,
                                                  targetDict,
                                                  nBreakTargets,
                                                  breakTimePerTarget)

        restroomTargets = generate_scenario_elements(1, 1,
                                                  largeRoomX + largeRoomWidth + wallThickness + smallRoomWidth / 2 - breakTargetWidth / 2,
                                                  stageLeftY + relativeTargetPositionY - chairRowSpacing,
                                                  breakTargetSpacingX + breakTargetWidth,
                                                  breakTargetSpacingY,
                                                  restroomTargetsFirstId,
                                                  breakTargetWidth,
                                                  breakTargetHeight,
                                                  targetDict,
                                                  nRestroomTargets,
                                                  breakTimePerTarget)

        finalTarget = [{"id": finalTargetId,
                        "absorbing": True,
                        "shape": {"x": largeRoomX - wallThickness,
                                  "y": boundingBoxWidth + wallThickness,
                                  "width": 0.5,
                                  "height": 2,
                                  "type": "RECTANGLE"},
                        "waitingTime": 0,
                        "waitingTimeYellowPhase": 0.0,
                        "parallelWaiters": 0,
                        "individualWaiting": True,
                        "deletionDistance": 0.1,
                        "startingWithRedLight": False,
                        "nextSpeed": -1.0}]

        targetList = stageLeftTargets + stageRightTargets + benchesTargets + breakTargets + restroomTargets + finalTarget

        # obstacles
        # room L and S, lower wall
        obstacleIdCounter = 2001
        entryWidth = 2
        obstacle1 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth,
                "y": boundingBoxWidth,
                "width": largeRoomWidth + smallRoomWidth + 3 * wallThickness,
                "height": wallThickness,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }

        obstacleIdCounter = obstacleIdCounter + 1

        # room L, left wall
        obstacle2 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth,
                "y": boundingBoxWidth + entryWidth + wallThickness,
                "width": wallThickness,
                "height": largeRoomHeight - entryWidth + wallThickness,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }
        obstacleIdCounter = obstacleIdCounter + 1

        # room L, upper wall
        obstacle3 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth + wallThickness,
                "y": boundingBoxWidth + largeRoomHeight + wallThickness,
                "width": largeRoomWidth + wallThickness + smallRoomWidth,
                "height": wallThickness,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }
        obstacleIdCounter = obstacleIdCounter + 1

        # wall between room L and S
        obstacle4 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth + wallThickness + largeRoomWidth,
                "y": boundingBoxWidth + entryWidth + wallThickness,
                "width": wallThickness,
                "height": largeRoomHeight - entryWidth * 2,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }
        obstacleIdCounter = obstacleIdCounter + 1

        # room S, right wall
        obstacle5 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth + wallThickness + largeRoomWidth + wallThickness + smallRoomWidth,
                "y": boundingBoxWidth + wallThickness,
                "width": wallThickness,
                "height": largeRoomHeight + wallThickness,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }
        obstacleIdCounter = obstacleIdCounter + 1

        # room S, right wall
        obstacle6 = {
            "shape": {
                "x": boundingBoxWidth + sourcesAreaWidth + wallThickness + largeRoomWidth + wallThickness,
                "y": benchesY + stageWallSpacingYUpper,
                "width": smallRoomWidth,
                "height": wallThickness,
                "type": "RECTANGLE"
            },
            "id": obstacleIdCounter
        }

        obstacleList = [obstacle1, obstacle2, obstacle3, obstacle4, obstacle5, obstacle6]

        # create new scenario file
        newScenarioName = get_current_scenario_name(scenario, seed)
        with open(pathToScenarioTemplateFile) as f:
            scenarioFile = json.load(f)

        # replace content in scenario file
        scenarioFile['scenario']['topography']['attributes']['boundingBoxWidth'] = boundingBoxWidth
        scenarioFile['scenario']['topography']['attributes']['bounds']['width'] = boundsX
        scenarioFile['scenario']['topography']['attributes']['bounds']['height'] = boundsY
        scenarioFile['scenario']['topography']['sources'] = sourceList
        scenarioFile['scenario']['topography']['targets'] = targetList
        scenarioFile['scenario']['topography']['obstacles'] = obstacleList
        scenarioFile['name'] = newScenarioName

        # save new scenario file
        pathToNewScenarioFile = os.path.join(pathToScenarioDir, newScenarioName + ".scenario")
        with open(pathToNewScenarioFile, 'w') as f:
            json.dump(scenarioFile, f, indent=2)

    if createNewTargetList:
        targetDataframe = pd.DataFrame.from_dict(targetDictionary)
        pd.to_pickle(targetDataframe, dumpedFilePath + '.pkl')
        targetDataframe.to_csv(dumpedFilePath + '.csv', index=False)
